#!/usr/bin/env python
"""Tabularizes static data in MEDS format into tabular representations."""

from itertools import product
from pathlib import Path

import hydra
import numpy as np
import polars as pl

pl.enable_string_cache()

from importlib.resources import files

from omegaconf import DictConfig

from ..describe_codes import (
    convert_to_df,
    filter_parquet,
    get_feature_columns,
    get_feature_freqs,
)
from ..file_name import list_subdir_files
from ..generate_static_features import get_flat_static_rep
from ..mapper import wrap as rwlock_wrap
from ..utils import (
    STATIC_CODE_AGGREGATION,
    STATIC_VALUE_AGGREGATION,
    filter_to_codes,
    get_shard_prefix,
    hydra_loguru_init,
    load_tqdm,
    stage_init,
    write_df,
)

config_yaml = files("MEDS_tabular_automl").joinpath("configs/tabularization.yaml")
if not config_yaml.is_file():
    raise FileNotFoundError("Core configuration not successfully installed!")


@hydra.main(version_base=None, config_path=str(config_yaml.parent.resolve()), config_name=config_yaml.stem)
def main(
    cfg: DictConfig,
):
    """Writes a flat (historically summarized) representation of the dataset to disk.

    This file caches a set of files useful for building flat representations of the dataset to disk,
    suitable for, e.g., sklearn style modeling for downstream tasks. It will produce a few sets of files:

    * A new directory ``self.config.save_dir / "flat_reps"`` which contains the following:
    * A subdirectory ``raw`` which contains: (1) a json file with the configuration arguments and (2) a
        set of parquet files containing flat (e.g., wide) representations of summarized events per subject,
        broken out by split and subject chunk.
    * A set of subdirectories ``past/*`` which contains summarized views over the past ``*`` time period
        per subject per event, for all time periods in ``window_sizes``, if any.

    Args:
        cfg:
            input_dir: directory of MEDS format dataset that is ingested.
            tabularized_data_dir: output directory of tabularized data.
            min_code_inclusion_count: The base feature inclusion count that should be used to dictate
                what features can be included in the flat representation. It can either be a float, in which
                case it applies across all measurements, or `None`, in which case no filtering is applied, or
                a dictionary from measurement type to a float dictating a per-measurement-type inclusion
                cutoff.
            window_sizes: Beyond writing out a raw, per-event flattened representation, the dataset also has
                the capability to summarize these flattened representations over the historical windows
                specified in this argument. These are strings specifying time deltas, using this syntax:
                `link`_. Each window size will be summarized to a separate directory, and will share the same
                subject file split as is used in the raw representation files.
            codes: A list of codes to include in the flat representation. If `None`, all codes will be included
                in the flat representation.
            aggs: A list of aggregations to apply to the raw representation. Must have length greater than 0.
            n_patients_per_sub_shard: The number of subjects that should be included in each output file.
                Lowering this number increases the number of files written, making the process of creating and
                leveraging these files slower but more memory efficient.
            do_overwrite: If `True`, this function will overwrite the data already stored in the target save
                directory.
            do_update: bool = True
            seed: The seed to use for random number generation.

    .. _link: https://pola-rs.github.io/polars/py-polars/html/reference/dataframe/api/polars.DataFrame.groupby_rolling.html # noqa: E501
    """
    stage_init(
        cfg,
        [
            "input_code_metadata_fp",
            "input_dir",
            "tabularization.filtered_code_metadata_fp",
        ],
    )
    iter_wrapper = load_tqdm(cfg.tqdm)
    if not cfg.loguru_init:
        hydra_loguru_init()

    # Step 1: Cache the filtered features that will be used in the tabularization process and modeling
    def read_fn(_):
        return _

    def compute_fn(_):
        filtered_feature_columns = filter_to_codes(
            cfg.input_code_metadata_fp,
            cfg.tabularization.allowed_codes,
            cfg.tabularization.min_code_inclusion_count,
            cfg.tabularization.min_code_inclusion_frequency,
            cfg.tabularization.max_included_codes,
        )
        feature_freqs = get_feature_freqs(cfg.input_code_metadata_fp)
        filtered_feature_columns_set = set(filtered_feature_columns)
        filtered_feature_freqs = {
            code: count for code, count in feature_freqs.items() if code in filtered_feature_columns_set
        }
        return convert_to_df(filtered_feature_freqs)

    def write_fn(data, out_fp):
        data.write_parquet(out_fp)

    in_fp = Path(cfg.input_code_metadata_fp)
    out_fp = Path(cfg.tabularization.filtered_code_metadata_fp)
    rwlock_wrap(
        in_fp,
        out_fp,
        read_fn,
        write_fn,
        compute_fn,
        do_overwrite=cfg.do_overwrite,
        do_return=False,
    )

    # Step 2: Produce static data representation
    meds_shard_fps = list_subdir_files(cfg.input_dir, "parquet")
    feature_columns = get_feature_columns(cfg.tabularization.filtered_code_metadata_fp)

    # shuffle tasks
    aggs = cfg.tabularization.aggs
    static_aggs = [agg for agg in aggs if agg in [STATIC_CODE_AGGREGATION, STATIC_VALUE_AGGREGATION]]
    tabularization_tasks = list(product(meds_shard_fps, static_aggs))
    np.random.shuffle(tabularization_tasks)
    for shard_fp, agg in iter_wrapper(tabularization_tasks):
        out_fp = (
            Path(cfg.output_tabularized_dir) / get_shard_prefix(cfg.input_dir, shard_fp) / "none" / agg
        ).with_suffix(".npz")
        if out_fp.exists() and not cfg.do_overwrite:
            raise FileExistsError(f"do_overwrite is {cfg.do_overwrite} and {out_fp} exists!")

        def read_fn(in_fp):
            return filter_parquet(in_fp, cfg.tabularization._resolved_codes)

        def compute_fn(shard_df):
            return get_flat_static_rep(
                agg=agg,
                feature_columns=feature_columns,
                shard_df=shard_df,
            )

        def write_fn(data, out_df):
            write_df(data, out_df, do_overwrite=cfg.do_overwrite)

        rwlock_wrap(
            shard_fp,
            out_fp,
            read_fn,
            write_fn,
            compute_fn,
            do_overwrite=cfg.do_overwrite,
            do_return=False,
        )


if __name__ == "__main__":
    main()



"""This module provides functions for generating static representations of patient data for use in automated
machine learning models. It includes functionality to summarize measurements based on static features and then
transform them into a tabular format suitable for analysis. The module leverages the polars library for
efficient data manipulation.

Functions:
- convert_to_matrix: Converts a Polars DataFrame to a sparse matrix.
- get_sparse_static_rep: Merges static and time-series dataframes into a sparse representation.
- summarize_static_measurements: Summarizes static measurements from a given DataFrame.
- get_flat_static_rep: Produces a tabular representation of static data features.
"""

import numpy as np
import polars as pl
from loguru import logger
from scipy.sparse import coo_array, csr_array

from MEDS_tabular_automl.utils import (
    STATIC_CODE_AGGREGATION,
    STATIC_VALUE_AGGREGATION,
    get_events_df,
    get_feature_names,
    get_unique_time_events_df,
    parse_static_feature_column,
)


def convert_to_matrix(df: pl.DataFrame, num_events: int, num_features: int) -> csr_array:
    """Converts a Polars DataFrame to a sparse matrix.

    Args:
        df: The DataFrame to convert.
        num_events: Number of events to set matrix dimension.
        num_features: Number of features to set matrix dimension.

    Returns:
        A sparse matrix representation of the DataFrame.
    """
    dense_matrix = df.drop("subject_id").collect().to_numpy()
    data_list = []
    rows = []
    cols = []
    for row in range(dense_matrix.shape[0]):
        for col in range(dense_matrix.shape[1]):
            data = dense_matrix[row, col]
            if (data is not None) and (data != 0):
                data_list.append(data)
                rows.append(row)
                cols.append(col)
    matrix = csr_array((data_list, (rows, cols)), shape=(num_events, num_features))
    return matrix


def get_sparse_static_rep(
    static_features: list[str], static_df: pl.DataFrame, meds_df: pl.DataFrame, feature_columns: list[str]
) -> coo_array:
    """Merges static and time-series dataframes into a sparse representation based on the subject_id column.

    Args:
        static_features: A list of static feature names.
        static_df: A DataFrame containing static features.
        meds_df: A DataFrame containing time-series features.
        feature_columns (list[str]): A list of feature columns to include in the merged DataFrame.

    Returns:
        A sparse array representation of the merged static and time-series features.
    """
    # Make static data sparse and merge it with the time-series data
    logger.info("Make static data sparse and merge it with the time-series data")
    # Check static_df is sorted and unique and raise error if it is not
    if not static_df.select(pl.col("subject_id")).collect().to_series().is_sorted():
        raise ValueError("static_df is not sorted by subject_id.")
    if not (
        static_df.select(pl.len()).collect().item()
        == static_df.select(pl.col("subject_id").n_unique()).collect().item()
    ):
        raise ValueError("static_df has duplicate subject_id values.")

    meds_df = get_unique_time_events_df(get_events_df(meds_df, feature_columns))

    # load static data as sparse matrix
    static_matrix = convert_to_matrix(
        static_df, num_events=meds_df.select(pl.len()).collect().item(), num_features=len(static_features)
    )
    # Duplicate static matrix rows to match time-series data
    events_per_patient = (
        meds_df.select(pl.col("subject_id").value_counts())
        .unnest("subject_id")
        .sort(by="subject_id")
        .select(pl.col("count"))
        .collect()
        .to_series()
    )
    reindex_slices = np.repeat(range(len(events_per_patient)), events_per_patient)
    static_matrix = static_matrix[reindex_slices, :]
    return coo_array(static_matrix)


def summarize_static_measurements(
    agg: str,
    feature_columns: list[str],
    df: pl.LazyFrame,
) -> pl.LazyFrame:
    """Aggregates static measurements for feature columns that are marked as 'present' or 'first'.

    This function first filters for features that need to be recorded as the first occurrence
    or simply as present, then performs a pivot to reshape the data for each patient, providing
    a tabular format where each row represents a patient and each column represents a static feature.

    Args:
        agg: The type of aggregation ('present' or 'first').
        feature_columns: A list of feature column identifiers marked for static analysis.
        df: The DataFrame from which features will be extracted and summarized.

    Returns:
        A LazyFrame containing summarized data pivoted by 'subject_id' for each static feature.
    """
    if agg == STATIC_VALUE_AGGREGATION:
        static_features = get_feature_names(agg=agg, feature_columns=feature_columns)
        # Handling 'first' static values
        static_first_codes = [parse_static_feature_column(c)[0] for c in static_features]
        code_subset = df.filter(pl.col("code").is_in(static_first_codes))
        first_code_subset = code_subset.group_by(pl.col("subject_id")).first().collect()
        static_value_pivot_df = first_code_subset.pivot(
            index=["subject_id"], columns=["code"], values=["numeric_value"], aggregate_function=None
        )
        # rename code to feature name
        remap_cols = {
            input_name: output_name
            for input_name, output_name in zip(static_first_codes, static_features)
            if input_name in static_value_pivot_df.columns
        }
        static_value_pivot_df = static_value_pivot_df.select(
            *["subject_id"], *[pl.col(k).alias(v).cast(pl.Boolean) for k, v in remap_cols.items()]
        ).sort(by="subject_id")
        # pivot can be faster: https://stackoverflow.com/questions/73522017/replacing-a-pivot-with-a-lazy-groupby-operation # noqa: E501
        # TODO: consider casting with .cast(pl.Float32))
        return static_value_pivot_df
    elif agg == STATIC_CODE_AGGREGATION:
        static_features = get_feature_names(agg=agg, feature_columns=feature_columns)
        # Handling 'present' static indicators
        static_present_codes = [parse_static_feature_column(c)[0] for c in static_features]
        static_present_pivot_df = (
            df.select(*["subject_id", "code"])
            .filter(pl.col("code").is_in(static_present_codes))
            .with_columns(pl.lit(True).alias("__indicator"))
            .collect()
            .pivot(
                index=["subject_id"],
                columns=["code"],
                values="__indicator",
                aggregate_function=None,
            )
            .sort(by="subject_id")
        )
        remap_cols = {
            input_name: output_name
            for input_name, output_name in zip(static_present_codes, static_features)
            if input_name in static_present_pivot_df.columns
        }
        # rename columns to final feature names
        static_present_pivot_df = static_present_pivot_df.select(
            *["subject_id"], *[pl.col(k).alias(v).cast(pl.Boolean) for k, v in remap_cols.items()]
        )
        return static_present_pivot_df
    else:
        raise ValueError(f"Invalid aggregation type: {agg}")


def get_flat_static_rep(
    agg: str,
    feature_columns: list[str],
    shard_df: pl.LazyFrame,
) -> coo_array:
    """Produces a sparse representation for static data from a specified shard DataFrame.

    This function selects the appropriate static features, summarizes them using
    summarize_static_measurements, and then normalizes the resulting data to ensure
    it is suitable for further analysis or machine learning tasks.

    Args:
        agg: The aggregation method for static data.
        feature_columns: A list of feature columns to include.
        shard_df: The shard DataFrame containing the patient data.

    Returns:
        A sparse array representing the static features for the provided shard of data.
    """
    static_features = get_feature_names(agg=agg, feature_columns=feature_columns)
    static_measurements = summarize_static_measurements(agg, static_features, df=shard_df)
    if len(static_features) == 0:
        raise ValueError(f"No static features found. Remove the aggregation function {agg}")
    # convert to sparse_matrix
    matrix = get_sparse_static_rep(static_features, static_measurements.lazy(), shard_df, feature_columns)
    if not matrix.shape[1] == len(static_features):
        raise ValueError(f"Expected {len(static_features)} features, got {matrix.shape[1]}")
    return matrix



"""Helper functions for getting file names and paths for MEDS tabular automl tasks."""
from pathlib import Path

from omegaconf import DictConfig


def list_subdir_files(root: Path | str, ext: str) -> list[Path]:
    """List files in subdirectories of a directory with a given extension.

    Args:
        root: Path to the directory.
        ext: File extension to filter files.

    Returns:
        An alphabetically sorted list of Path objects to files matching the extension in any level of
        subdirectories of the given directory.

    Examples:
        >>> import tempfile
        >>> tmpdir = tempfile.TemporaryDirectory()
        >>> root = Path(tmpdir.name)
        >>> subdir_1 = root / "subdir_1"
        >>> subdir_1.mkdir()
        >>> subdir_2 = root / "subdir_2"
        >>> subdir_2.mkdir()
        >>> subdir_1_A = subdir_1 / "A"
        >>> subdir_1_A.mkdir()
        >>> (root / "1.csv").touch()
        >>> (root / "foo.parquet").touch()
        >>> (root / "2.csv").touch()
        >>> (root / "subdir_1" / "3.csv").touch()
        >>> (root / "subdir_2" / "4.csv").touch()
        >>> (root / "subdir_1" / "A" / "5.csv").touch()
        >>> (root / "subdir_1" / "A" / "15.csv.gz").touch()
        >>> [fp.relative_to(root) for fp in list_subdir_files(root, "csv")] # doctest: +NORMALIZE_WHITESPACE
        [PosixPath('1.csv'),
         PosixPath('2.csv'),
         PosixPath('subdir_1/3.csv'),
         PosixPath('subdir_1/A/5.csv'),
         PosixPath('subdir_2/4.csv')]
        >>> [fp.relative_to(root) for fp in list_subdir_files(root, "parquet")]
        [PosixPath('foo.parquet')]
        >>> [fp.relative_to(root) for fp in list_subdir_files(root, "csv.gz")]
        [PosixPath('subdir_1/A/15.csv.gz')]
        >>> [fp.relative_to(root) for fp in list_subdir_files(root, "json")]
        []
        >>> list_subdir_files(root / "nonexistent", "csv")
        []
        >>> tmpdir.cleanup()
    """

    return sorted(list(Path(root).glob(f"**/*.{ext}")))


def get_model_files(cfg: DictConfig, split: str, shard: str) -> list[Path]:
    """Get the tabularized npz files for a given split and shard number.

    TODO: Rename function to get_tabularized_input_files or something.

    Args:
        cfg: `OmegaConf.DictConfig` object with the configuration. It must have the following keys:
            - input_dir: Path to the directory with the tabularized npz files.
            - tabularization: Tabularization configuration, as a nested `DictConfig` object with keys:
              - window_sizes: List of window sizes.
              - aggs: List of aggregation functions.
        split: Split name to reference the files stored on disk.
        shard: The shard within the split to reference the files stored on disk.

    Returns:
        An alphabetically sorted list of Path objects to the tabularized npz files for the given split and
        shard. These files will take the form ``{cfg.input_dir}/{split}/{shard}/{window_size}/{agg}.npz``. For
        static aggregations, the window size will be "none" as these features are not time-varying.

    Examples:
        >>> cfg = DictConfig({
        ...     "path": DictConfig({"input_tabularized_cache_dir" : "data"}),
        ...     "tabularization": {
        ...         "window_sizes": ["1d", "7d"],
        ...         "aggs": ["code/count", "value/sum", "static/present"],
        ...     }
        ... })
        >>> get_model_files(cfg, "train", "0") # doctest: +NORMALIZE_WHITESPACE
        [PosixPath('data/train/0/1d/code/count.npz'),
         PosixPath('data/train/0/1d/value/sum.npz'),
         PosixPath('data/train/0/7d/code/count.npz'),
         PosixPath('data/train/0/7d/value/sum.npz'),
         PosixPath('data/train/0/none/static/present.npz')]
        >>> get_model_files(cfg, "test/IID", "3/0") # doctest: +NORMALIZE_WHITESPACE
        [PosixPath('data/test/IID/3/0/1d/code/count.npz'),
         PosixPath('data/test/IID/3/0/1d/value/sum.npz'),
         PosixPath('data/test/IID/3/0/7d/code/count.npz'),
         PosixPath('data/test/IID/3/0/7d/value/sum.npz'),
         PosixPath('data/test/IID/3/0/none/static/present.npz')]
    """
    window_sizes = cfg.tabularization.window_sizes
    aggs = cfg.tabularization.aggs
    shard_dir = Path(cfg.path.input_tabularized_cache_dir) / split / shard
    # Given a shard number, returns the model files
    model_files = []
    for window_size in window_sizes:
        for agg in aggs:
            if agg.startswith("static"):
                continue
            else:
                model_files.append(shard_dir / window_size / f"{agg}.npz")
    for agg in aggs:
        if agg.startswith("static"):
            window_size = "none"
            model_files.append(shard_dir / window_size / f"{agg}.npz")
    return sorted(model_files)



from pathlib import Path

import polars as pl

from MEDS_tabular_automl.utils import get_feature_names


def convert_to_df(freq_dict: dict[str, int]) -> pl.DataFrame:
    """Converts a dictionary of code frequencies to a Polars DataFrame.

    Args:
        freq_dict: A dictionary with code features and their respective frequencies.

    Returns:
        A DataFrame with two columns, "code" and "count".

    TODOs:
        - Eliminate this function and just use a DataFrame throughout. See #14
        - Use categorical types for `code` instead of strings.

    Examples:
        >>> convert_to_df({"A": 1, "B": 2, "C": 3})
        shape: (3, 2)
        ┌──────┬───────┐
        │ code ┆ count │
        │ ---  ┆ ---   │
        │ str  ┆ i64   │
        ╞══════╪═══════╡
        │ A    ┆ 1     │
        │ B    ┆ 2     │
        │ C    ┆ 3     │
        └──────┴───────┘
    """
    return pl.DataFrame([[col, freq] for col, freq in freq_dict.items()], schema=["code", "count"])


def convert_to_freq_dict(df: pl.LazyFrame) -> dict[str, dict[int, int]]:
    """Converts a DataFrame to a dictionary of frequencies.

    Args:
        df: The DataFrame to be converted.

    Returns:
        A dictionary where keys are column names and values are
        dictionaries of code frequencies.

    Raises:
        ValueError: If the DataFrame does not have the expected columns "code" and "count".

    TODOs:
        - Eliminate this function and just use a DataFrame throughout. See #14

    Example:
        >>> import polars as pl
        >>> data = pl.DataFrame({"code": [1, 2, 3, 4, 5], "count": [10, 20, 30, 40, 50]}).lazy()
        >>> convert_to_freq_dict(data)
        {1: 10, 2: 20, 3: 30, 4: 40, 5: 50}
        >>> convert_to_freq_dict(pl.DataFrame({"code": ["A", "B", "C"], "value": [1, 2, 3]}).lazy())
        Traceback (most recent call last):
            ...
        ValueError: DataFrame must have columns 'code' and 'count', but has columns ['code', 'value']!
    """
    if not df.columns == ["code", "count"]:
        raise ValueError(f"DataFrame must have columns 'code' and 'count', but has columns {df.columns}!")
    return dict(df.collect().iter_rows())


def compute_feature_frequencies(shard_df: pl.LazyFrame) -> pl.DataFrame:
    """Generates a DataFrame containing the frequencies of codes and numerical values under different
    aggregations by computing frequency counts for certain attributes and organizing the results into specific
    categories based on the dataset's features.

    Args:
        shard_df: A DataFrame containing the data to be analyzed and split (e.g., 'train', 'test').

    Returns:
        A tuple containing a list of feature columns and a dictionary of code properties identified
        during the evaluation.

    Examples:
        >>> from datetime import datetime
        >>> data = pl.DataFrame({
        ...     'subject_id': [1, 1, 2, 2, 3, 3, 3],
        ...     'code': ['A', 'A', 'B', 'B', 'C', 'C', 'C'],
        ...     'time': [
        ...         None,
        ...         datetime(2021, 1, 1),
        ...         None,
        ...         None,
        ...         datetime(2021, 1, 3),
        ...         datetime(2021, 1, 4),
        ...         None
        ...     ],
        ...     'numeric_value': [1, None, 2, 2, None, None, 3]
        ... }).lazy()
        >>> assert (
        ...     convert_to_freq_dict(compute_feature_frequencies(data).lazy()) == {
        ...         'B/static/present': 2, 'C/static/present': 1, 'A/static/present': 1, 'B/static/first': 2,
        ...         'C/static/first': 1, 'A/static/first': 1, 'A/code': 1, 'C/code': 2
        ...     }
        ... )
    """
    static_df = shard_df.filter(
        pl.col("subject_id").is_not_null() & pl.col("code").is_not_null() & pl.col("time").is_null()
    )
    static_code_freqs_df = static_df.group_by("code").agg(pl.count("code").alias("count")).collect()
    static_code_freqs = {
        row["code"] + "/static/present": row["count"] for row in static_code_freqs_df.iter_rows(named=True)
    }

    static_value_df = static_df.filter(pl.col("numeric_value").is_not_null())
    static_value_freqs_df = (
        static_value_df.group_by("code").agg(pl.count("numeric_value").alias("count")).collect()
    )
    static_value_freqs = {
        row["code"] + "/static/first": row["count"] for row in static_value_freqs_df.iter_rows(named=True)
    }

    ts_df = shard_df.filter(
        pl.col("subject_id").is_not_null() & pl.col("code").is_not_null() & pl.col("time").is_not_null()
    )
    code_freqs_df = ts_df.group_by("code").agg(pl.count("code").alias("count")).collect()
    code_freqs = {row["code"] + "/code": row["count"] for row in code_freqs_df.iter_rows(named=True)}

    value_df = ts_df.filter(pl.col("numeric_value").is_not_null())
    value_freqs_df = value_df.group_by("code").agg(pl.count("numeric_value").alias("count")).collect()
    value_freqs = {row["code"] + "/value": row["count"] for row in value_freqs_df.iter_rows(named=True)}

    combined_freqs = {**static_code_freqs, **static_value_freqs, **code_freqs, **value_freqs}
    return convert_to_df(combined_freqs)


def get_feature_columns(fp: Path) -> list[str]:
    """Retrieves feature column names from a parquet file.

    Args:
        fp: File path to the Parquet data.

    Returns:
        Sorted list of column names.

    Examples:
        >>> from tempfile import NamedTemporaryFile
        >>> with NamedTemporaryFile() as f:
        ...     pl.DataFrame({"code": ["E", "D", "A"], "count": [1, 3, 2]}).write_parquet(f.name)
        ...     get_feature_columns(f.name)
        ['A', 'D', 'E']
    """
    return sorted(list(get_feature_freqs(fp).keys()))


def get_feature_freqs(fp: Path) -> dict[str, int]:
    """Retrieves feature frequencies from a parquet file.

    Args:
        fp: File path to the Parquet data.

    Returns:
        Dictionary of feature frequencies.

    Examples:
        >>> from tempfile import NamedTemporaryFile
        >>> with NamedTemporaryFile() as f:
        ...     pl.DataFrame({"code": ["E", "D", "A"], "count": [1, 3, 2]}).write_parquet(f.name)
        ...     get_feature_freqs(f.name)
        {'E': 1, 'D': 3, 'A': 2}
    """
    return convert_to_freq_dict(pl.scan_parquet(fp))


def clear_code_aggregation_suffix(code: str) -> str:
    """Removes aggregation suffixes from code strings.

    Args:
        code: Code string to be cleared.

    Returns:
        Code string without aggregation suffixes.

    Raises:
        ValueError: If the code does not have a recognized aggregation suffix.

    Examples:
        >>> clear_code_aggregation_suffix("A/code")
        'A'
        >>> clear_code_aggregation_suffix("A/value")
        'A'
        >>> clear_code_aggregation_suffix("A/static/present")
        'A'
        >>> clear_code_aggregation_suffix("A/static/first")
        'A'
        >>> clear_code_aggregation_suffix("A")
        Traceback (most recent call last):
            ...
        ValueError: Code A does not have a recognized aggregation suffix!
    """
    if code.endswith("/code"):
        return code[:-5]
    elif code.endswith("/value"):
        return code[:-6]
    elif code.endswith("/static/present"):
        return code[:-15]
    elif code.endswith("/static/first"):
        return code[:-13]
    else:
        raise ValueError(f"Code {code} does not have a recognized aggregation suffix!")


def filter_parquet(fp: Path, allowed_codes: list[str]) -> pl.LazyFrame:
    """Loads and filters a Parquet file with Polars to include only specified codes and removes rare
    codes/values.

    Args:
        fp: Path to the Parquet file of a Meds cohort shard.
        allowed_codes: List of codes to filter by.

    Returns:
        pl.LazyFrame: A filtered LazyFrame containing only the allowed and not rare codes/values.

    Examples:
        >>> from tempfile import NamedTemporaryFile
        >>> fp = NamedTemporaryFile()
        >>> pl.DataFrame({
        ...     "code": ["A", "A", "A", "A", "D", "D", "E", "E"],
        ...     "time": [None, None, "2021-01-01", "2021-01-01", None, None, "2021-01-03", "2021-01-04"],
        ...     "numeric_value": [1, None, 2, 2, None, 5, None, 3]
        ... }).write_parquet(fp.name)
        >>> filter_parquet(fp.name, ["A/code", "D/static/present", "E/code", "E/value"]).collect()
        shape: (6, 3)
        ┌──────┬────────────┬───────────────┐
        │ code ┆ time       ┆ numeric_value │
        │ ---  ┆ ---        ┆ ---           │
        │ str  ┆ str        ┆ i64           │
        ╞══════╪════════════╪═══════════════╡
        │ A    ┆ 2021-01-01 ┆ null          │
        │ A    ┆ 2021-01-01 ┆ null          │
        │ D    ┆ null       ┆ null          │
        │ D    ┆ null       ┆ null          │
        │ E    ┆ 2021-01-03 ┆ null          │
        │ E    ┆ 2021-01-04 ┆ 3             │
        └──────┴────────────┴───────────────┘
        >>> fp.close()
    """
    df = pl.scan_parquet(fp)
    # Drop values that are rare
    # Drop Rare Static Codes
    static_present_feature_columns = [
        clear_code_aggregation_suffix(each) for each in get_feature_names("static/present", allowed_codes)
    ]
    static_first_feature_columns = [
        clear_code_aggregation_suffix(each) for each in get_feature_names("static/first", allowed_codes)
    ]
    code_feature_columns = [
        clear_code_aggregation_suffix(each) for each in get_feature_names("code/count", allowed_codes)
    ]
    value_feature_columns = [
        clear_code_aggregation_suffix(each) for each in get_feature_names("value/sum", allowed_codes)
    ]

    is_static_code = pl.col("time").is_null()
    is_numeric_code = pl.col("numeric_value").is_not_null()
    rare_static_code = is_static_code & ~pl.col("code").is_in(static_present_feature_columns)
    rare_ts_code = ~is_static_code & ~pl.col("code").is_in(code_feature_columns)
    rare_ts_value = ~is_static_code & ~pl.col("code").is_in(value_feature_columns) & is_numeric_code
    rare_static_value = is_static_code & ~pl.col("code").is_in(static_first_feature_columns) & is_numeric_code

    # Remove rare numeric values by converting them to null
    df = df.with_columns(
        pl.when(rare_static_value | rare_ts_value)
        .then(None)
        .otherwise(pl.col("numeric_value"))
        .alias("numeric_value")
    )
    # Drop rows with rare codes
    df = df.filter(~(rare_static_code | rare_ts_code))
    return df



import shutil
from pathlib import Path

import polars as pl
from hydra.experimental.callback import Callback
from loguru import logger
from omegaconf import DictConfig


class EvaluationCallback(Callback):
    def on_multirun_end(self, config: DictConfig, **kwargs):
        """Find best model based on log files and logger.info its performance and hyperparameters."""
        log_fp = Path(config.path.sweep_results_dir)

        try:
            performance = pl.read_csv(log_fp / f"*/*{config.path.performance_log_stem}.log")
        except Exception as e:
            raise FileNotFoundError(f"Log files incomplete or not found at {log_fp}") from e

        performance = performance.sort("tuning_auc", descending=True, nulls_last=True)
        logger.info(f"\nPerformance of the top 10 models:\n{performance.head(10)}")

        self.log_performance(performance[0, :])
        if hasattr(config, "delete_below_top_k") and config.delete_below_top_k >= 0:
            self.delete_below_top_k_models(
                performance, config.delete_below_top_k, config.path.sweep_results_dir
            )
        else:
            logger.info(
                "All models were saved. To automatically delete models, set delete_below_top_k in config."
            )
        best_trial_dir = Path(config.path.sweep_results_dir) / performance["trial_name"].cast(pl.String)[0]
        output_best_trial_dir = Path(config.path.best_trial_dir)
        shutil.copytree(best_trial_dir, output_best_trial_dir)
        performance.write_parquet(config.time_output_model_dir / "sweep_results_summary.parquet")

        return performance.head(1)

    def log_performance(self, best_model_performance):
        """logger.info performance of the best model with nice formatting."""
        best_model = best_model_performance["trial_name"][0]
        tuning_auc = best_model_performance["tuning_auc"][0]
        test_auc = best_model_performance["test_auc"][0]
        log_performance_message = [
            f"\nBest model can be found at {best_model}",
            "Performance of best model:",
            f"Tuning AUC: {tuning_auc}",
            f"Test AUC: {test_auc}",
        ]
        logger.info("\n".join(log_performance_message))

    def delete_below_top_k_models(self, performance, k, sweep_results_dir):
        """Save only top k models from the sweep results directory and delete all other directories.

        Args:
            performance: DataFrame containing trial_name and performance metrics.
            k: Number of top models to save.
            sweep_results_dir: Directory containing trial results.

        Example:
            >>> import tempfile
            >>> import json
            >>> import polars as pl
            >>> from pathlib import Path
            >>> performance = pl.DataFrame(
            ...     {
            ...         "trial_name": ["trial1", "trial2", "trial3", "trial4"],
            ...         "tuning_auc": [0.9, 0.8, 0.7, 0.6],
            ...         "test_auc": [0.9, 0.8, 0.7, 0.6],
            ...     }
            ... )
            >>> k = 2
            >>> with tempfile.TemporaryDirectory() as sweep_dir:
            ...     for trial in performance["trial_name"]:
            ...         trial_dir = Path(sweep_dir) / trial
            ...         trial_dir.mkdir()
            ...         with open(trial_dir / "model.json", 'w') as f:
            ...             json.dump({"model_name": trial, "content": "dummy data"}, f)
            ...     cb = EvaluationCallback()
            ...     cb.delete_below_top_k_models(performance, k, sweep_dir)
            ...     remaining_trials = sorted(p.name for p in Path(sweep_dir).iterdir())
            >>> remaining_trials
            ['trial1', 'trial2']
        """
        logger.info(f"Deleting all models except top {k} models.")
        top_k_models = performance.head(k)["trial_name"].cast(pl.String).to_list()
        logger.debug(f"Top {k} models: {top_k_models}")
        for trial_dir in Path(sweep_results_dir).iterdir():
            if trial_dir.stem not in top_k_models:
                shutil.rmtree(trial_dir)



#!/usr/bin/env python

"""Aggregates time-series data for feature columns across different window sizes."""
from importlib.resources import files
from pathlib import Path

import hydra
import numpy as np
import polars as pl
import scipy.sparse as sp
from loguru import logger
from omegaconf import DictConfig

from ..describe_codes import filter_parquet, get_feature_columns
from ..file_name import list_subdir_files
from ..mapper import wrap as rwlock_wrap
from ..utils import (
    CODE_AGGREGATIONS,
    STATIC_CODE_AGGREGATION,
    STATIC_VALUE_AGGREGATION,
    VALUE_AGGREGATIONS,
    get_events_df,
    get_shard_prefix,
    get_unique_time_events_df,
    hydra_loguru_init,
    load_matrix,
    load_tqdm,
    stage_init,
    write_df,
)

config_yaml = files("MEDS_tabular_automl").joinpath("configs/task_specific_caching.yaml")
if not config_yaml.is_file():
    raise FileNotFoundError("Core configuration not successfully installed!")


VALID_AGGREGATIONS = [
    *VALUE_AGGREGATIONS,
    *CODE_AGGREGATIONS,
    STATIC_CODE_AGGREGATION,
    STATIC_VALUE_AGGREGATION,
]


def write_lazyframe(df: pl.LazyFrame, fp: Path):
    df.collect().write_parquet(fp, use_pyarrow=True)


def generate_row_cached_matrix(matrix: sp.coo_array, label_df: pl.LazyFrame) -> sp.coo_array:
    """Generates row-cached matrix for a given matrix and label DataFrame.

    Args:
        matrix: The input sparse matrix.
        label_df: A LazyFrame with an 'event_id' column indicating valid row indices in the matrix.

    Returns:
        A COOrdinate formatted sparse matrix containing only the rows specified by label_df's event_ids.

    Raises:
        ValueError: If the maximum event_id in label_df exceeds the number of rows in the matrix.
    """
    label_len = label_df.select(pl.col("event_id").max()).collect().item()
    if matrix.shape[0] <= label_len:
        raise ValueError(
            f"Label_df event_ids must be valid indexes of sparse matrix: {matrix.shape[0]} <= {label_len}"
        )
    csr = sp.csr_array(matrix)
    valid_ids = label_df.select(pl.col("event_id")).collect().to_series().to_numpy()
    csr = csr[valid_ids, :]
    return sp.coo_array(csr)


@hydra.main(version_base=None, config_path=str(config_yaml.parent.resolve()), config_name=config_yaml.stem)
def main(cfg: DictConfig):
    """Performs row splicing of tabularized data for a specific task based on configuration.

    Uses Hydra to manage configurations and logging. The function processes data files based on specified
    task configurations, loading matrices, applying transformations, and writing results.

    Args:
        cfg: The configuration for processing, loaded from a YAML file.
    """
    stage_init(
        cfg,
        [
            "input_dir",
            "input_label_dir",
            "input_tabularized_dir",
            "output_dir",
            "tabularization.filtered_code_metadata_fp",
        ],
    )
    iter_wrapper = load_tqdm(cfg.tqdm)
    if not cfg.loguru_init:
        hydra_loguru_init()
    # Produce ts representation

    # shuffle tasks
    tabularization_tasks = list_subdir_files(cfg.input_tabularized_dir, "npz")
    if len(tabularization_tasks) == 0:
        raise FileNotFoundError(
            f"No tabularized data found, `input_tabularized_dir`: {cfg.input_tabularized_dir}, "
            "is likely incorrect"
        )

    np.random.shuffle(tabularization_tasks)

    label_dir = Path(cfg.input_label_dir)
    if not label_dir.exists():
        raise FileNotFoundError(
            f"Label directory {label_dir} does not exist, please check the `input_label_dir` kwarg"
        )
    label_df = (
        pl.scan_parquet(label_dir / "**/*.parquet")
        .rename(
            {
                "prediction_time": "time",
                cfg.label_column: "label",
            }
        )
        .group_by(pl.col("subject_id", "time"), maintain_order=True)
        .first()
    )

    feature_columns = get_feature_columns(cfg.tabularization.filtered_code_metadata_fp)

    # iterate through them
    for data_fp in iter_wrapper(tabularization_tasks):
        # parse as time series agg
        split, shard_num, window_size, code_type, agg_name = Path(data_fp).with_suffix("").parts[-5:]
        meds_data_in_fp = Path(cfg.input_dir) / split / f"{shard_num}.parquet"
        shard_label_fp = Path(cfg.output_label_cache_dir) / split / f"{shard_num}.parquet"
        out_fp = (
            Path(cfg.output_tabularized_cache_dir) / get_shard_prefix(cfg.input_tabularized_dir, data_fp)
        ).with_suffix(".npz")

        def read_meds_data_df(meds_data_fp):
            if "numeric_value" not in pl.scan_parquet(meds_data_fp).columns:
                raise ValueError(
                    f"'numeric_value' column not found in raw data {meds_data_fp}. "
                    "You are maybe loading labels instead of meds data"
                )
            return filter_parquet(meds_data_fp, cfg.tabularization._resolved_codes)

        def extract_labels(meds_data_df):
            meds_data_df = (
                get_unique_time_events_df(get_events_df(meds_data_df, feature_columns))
                .with_row_index("event_id")
                .select("subject_id", "time", "event_id")
            )
            shard_label_df = label_df.join(
                meds_data_df.select("subject_id").unique(), on="subject_id", how="inner"
            ).join_asof(other=meds_data_df, by="subject_id", on="time")
            return shard_label_df

        def read_fn(in_fp_tuple):
            meds_data_fp, data_fp = in_fp_tuple
            # TODO: replace this with more intelligent locking
            if not Path(shard_label_fp).exists():
                logger.info(f"Extracting labels for {shard_label_fp}")
                Path(shard_label_fp).parent.mkdir(parents=True, exist_ok=True)
                meds_data_df = read_meds_data_df(meds_data_fp)
                extracted_events = extract_labels(meds_data_df)
                write_lazyframe(extracted_events, shard_label_fp)
            else:
                logger.info(f"Labels already exist, reading from {shard_label_fp}")
            shard_label_df = pl.scan_parquet(shard_label_fp)
            matrix = load_matrix(data_fp)
            return shard_label_df, matrix

        def compute_fn(input_tuple):
            shard_label_df, matrix = input_tuple
            row_cached_matrix = generate_row_cached_matrix(matrix=matrix, label_df=shard_label_df)
            return row_cached_matrix

        def write_fn(row_cached_matrix, out_fp):
            write_df(row_cached_matrix, out_fp, do_overwrite=cfg.do_overwrite)

        rwlock_wrap(
            (meds_data_in_fp, data_fp),
            out_fp,
            read_fn,
            write_fn,
            compute_fn,
            do_overwrite=cfg.do_overwrite,
            do_return=False,
        )


if __name__ == "__main__":
    main()



import os
import shutil
import sys
from pathlib import Path

import MEDS_tabular_automl

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "MEDS-Tab"
copyright = "2024, Nassim Oufattole, Matthew McDermott, Teya Bergamaschi, Aleksia Kolo, Hyewon Jeong"
author = "Nassim Oufattole, Matthew McDermott, Teya Bergamaschi, Aleksia Kolo, Hyewon Jeong"
# Define the json_url for our version switcher.


json_url = "https://meds-tab.readthedocs.io/en/latest/_static/switcher.json"
# Define the version we use for matching in the version switcher.
version_match = os.environ.get("READTHEDOCS_VERSION")
release = MEDS_tabular_automl.__version__
# If READTHEDOCS_VERSION doesn't exist, we're not on RTD
# If it is an integer, we're in a PR build and the version isn't correct.
# If it's "latest" → change to "dev" (that's what we want the switcher to call it)
if not version_match or version_match.isdigit():
    # For local development, infer the version to match from the package.
    if "dev" in release or "rc" in release:
        version_match = "dev"
        # We want to keep the relative reference if we are in dev mode
        # but we want the whole url if we are effectively in a released version
        json_url = "_static/switcher.json"
    else:
        version_match = f"v{release}"
elif version_match == "latest":
    version_match = "dev"
elif version_match == "stable":
    version_match = f"v{release}"

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

language = "en"
# -- Path setup

__location__ = Path(os.path.dirname(__file__))
__src__ = __location__ / "../.."

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, str(__src__))


def ensure_pandoc_installed(_):
    """Source: https://stackoverflow.com/questions/62398231/building-docs-fails-due-to-missing-pandoc"""
    import pypandoc

    # Download pandoc if necessary. If pandoc is already installed and on
    # the PATH, the installed version will be used. Otherwise, we will
    # download a copy of pandoc into docs/bin/ and add that to our PATH.
    pandoc_dir = str(__location__ / "bin")
    # Add dir containing pandoc binary to the PATH environment variable
    if pandoc_dir not in os.environ["PATH"].split(os.pathsep):
        os.environ["PATH"] += os.pathsep + pandoc_dir

    pypandoc.ensure_pandoc_installed(
        targetfolder=pandoc_dir,
        delete_installer=True,
    )


# -- Run sphinx-apidoc
# This ensures we don't need to run apidoc manually.

# TODO: use https://github.com/sphinx-extensions2/sphinx-autodoc2

from sphinx.ext import apidoc

output_dir = __location__ / "api"
module_dir = __src__ / "src/MEDS_tabular_automl"
if output_dir.is_dir():
    shutil.rmtree(output_dir)

try:
    cmd_line = f"--implicit-namespaces -e -f -o {output_dir} {module_dir}"
    apidoc.main(cmd_line.split(" "))
except Exception as e:  # pylint: disable=broad-except
    print(f"Running `sphinx-apidoc {cmd_line}` failed!\n{e}")


# -- General configuration


# -- Project information
extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.ifconfig",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.imgconverter",
    "sphinxcontrib.collections",
    "sphinx_subfigure",
    "myst_parser",
    "nbsphinx",
    # "sphinx_immaterial",
]
# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "pydata_sphinx_theme"
# html_sidebars = {"**": []}  # ["logo-text.html", "globaltoc.html", "localtoc.html", "searchbox.html"]
html_sidebars = {
    "api/*": [
        "sidebar-nav-bs",
    ],
    "**": [],
}
nbsphinx_allow_errors = True


collections_dir = __location__ / "_collections"
if not collections_dir.is_dir():
    os.mkdir(collections_dir)

python_version = ".".join(map(str, sys.version_info[0:2]))
intersphinx_mapping = {
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "python": ("https://docs.python.org/" + python_version, None),
    "matplotlib": ("https://matplotlib.org", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "sklearn": ("https://scikit-learn.org/stable", None),
    "pandas": ("https://pandas.pydata.org/docs", None),
    "pandera": ("https://pandera.readthedocs.io/en/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "setuptools": ("https://setuptools.pypa.io/en/stable/", None),
    "pyscaffold": ("https://pyscaffold.org/en/stable", None),
    "hyperimpute": ("https://hyperimpute.readthedocs.io/en/latest/", None),
    "xgbse": ("https://loft-br.github.io/xgboost-survival-embeddings/", None),
    "lifelines": ("https://lifelines.readthedocs.io/en/stable/", None),
    "optuna": ("https://optuna.readthedocs.io/en/stable/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Control options for included jupyter notebooks.
nb_execution_mode = "off"


# -- Options for HTML output

# Configure MyST-Parser
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "substitution",
    "tasklist",
]

myst_update_mathjax = True

# MyST URL schemes.
myst_url_schemes = {
    "http": None,
    "https": None,
    "ftp": None,
    "mailto": None,
    "repo-code": "https://github.com/mmcdermott/MEDS_Tabular_AutoML/tree/main/{{path}}#{{fragment}}",
    # "doi": "https://doi.org/{{path}}",
    # "gh-issue": {
    #     "url": "https://github.com/executablebooks/MyST-Parser/issue/{{path}}#{{fragment}}",
    #     "title": "Issue #{{path}}",
    #     "classes": ["github"],
    # },
}

# The suffix of source filenames.
source_suffix = [".rst", ".md"]

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = "py:obj"

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
# https://pygments.org/styles/
pygments_style = "tango"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If this is True, todo emits a warning for each TODO entries. The default is False.
todo_emit_warnings = True


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.


html_title = f"MEDS-Tab v{release} Documentation"
html_short_title = "MEDS-Tab"

# html_logo = "query-512.png"
# html_favicon = "query-16.ico"

# Material theme options (see theme.conf for more information)
html_theme_options = {
    "logo": {
        "text": "MEDS-TAB",
        "image_light": "../assets/dark_purple_meds_tab.png",
        "image_dark": "../assets/light_purple_meds_tab.png",
    },
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/mmcdermott/MEDS_Tabular_AutoML",  # required
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/meds-tab/",
            "icon": "fa-brands fa-python",
        },
    ],
    "header_links_before_dropdown": 6,
    "show_toc_level": 1,
    "navbar_align": "left",  # [left, content, right] For testing that the navbar items align properly
    # "show_nav_level": 2,
    "announcement": "This is a community-supported tool. If you'd like to contribute, <a href='https://github.com/mmcdermott/MEDS_Tabular_AutoML'>check out our GitHub repository.</a> Your contributions are welcome!",  # noqa E501
    "show_version_warning_banner": True,
    "switcher": {
        "json_url": json_url,
        "version_match": version_match,
    },
    "navbar_center": ["version-switcher", "navbar-nav"],
    "footer_start": ["copyright"],
    "footer_center": ["sphinx-version"],
    "use_edit_page_button": True,
    # "secondary_sidebar_items": {
    #     "**/*": ["page-toc", "edit-this-page", "sourcelink"],
    # },
    "back_to_top_button": True,
}

html_context = {
    "github_user": "mmcdermott",
    "github_repo": "MEDS_Tabular_AutoML",
    "github_version": "main",
    "doc_path": "docs/source",
}

# html_theme_options = {
#     # Set the name of the project to appear in the navigation.
#     "nav_title": "MEDS-TAB",
#     "palette": {"primary": "purple", "accent": "purple"},
#     # {
#     #     "media": "(prefers-color-scheme: light)",
#     #     "scheme": "default",
#     #     "toggle": {
#     #         "icon": "material/toggle-switch-off-outline",
#     #         "name": "Switch to dark mode",
#     #     },
#     # },
#     # {
#     #     "media": "(prefers-color-scheme: dark)",
#     #     "scheme": "slate",
#     #     "toggle": {
#     #         "icon": "material/toggle-switch",
#     #         "name": "Switch to light mode",
#     #     },
#     # },
#     # "color_primary": "green",
#     # "color_accent": "green",
#     # Set the repo location to get a badge with stats
#     "repo_url": "https://github.com/mmcdermott/MEDS_Tabular_AutoML",
#     "repo_name": "meds-tab",
#     # Visible levels of the global TOC; -1 means unlimited
#     "globaltoc_depth": 3,
#     # If False, expand all TOC entries
#     "globaltoc_collapse": True,
#     # If True, show hidden TOC entries
#     "globaltoc_includehidden": False,
# }


html_show_copyright = True
htmlhelp_basename = "meds-tab-doc"

# -- Options for LaTeX output
# latex_engine = "xelatex"
latex_elements = {  # type: ignore
    # The paper size ("letterpaper" or "a4paper").
    "papersize": "letterpaper",
    # The font size ("10pt", "11pt" or "12pt").
    "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    "preamble": "\n".join(
        [
            r"\usepackage{svg}",
            r"\DeclareUnicodeCharacter{2501}{-}",
            r"\DeclareUnicodeCharacter{2503}{|}",
            r"\DeclareUnicodeCharacter{2500}{-}",
            r"\DeclareUnicodeCharacter{2550}{-}",
            r"\DeclareUnicodeCharacter{2517}{+}",
            r"\DeclareUnicodeCharacter{2518}{+}",
            r"\DeclareUnicodeCharacter{2534}{+}",
            r"\DeclareUnicodeCharacter{250C}{+}",
            r"\DeclareUnicodeCharacter{252C}{+}",
            r"\DeclareUnicodeCharacter{2510}{+}",
            r"\DeclareUnicodeCharacter{2502}{|}",
            r"\DeclareUnicodeCharacter{2506}{|}",
            r"\DeclareUnicodeCharacter{2561}{|}",
            r"\DeclareUnicodeCharacter{256A}{|}",
            r"\DeclareUnicodeCharacter{2523}{|}",
            r"\DeclareUnicodeCharacter{03BC}{\ensuremath{\mu}}",
            r"\DeclareUnicodeCharacter{255E}{|}",
            r"\DeclareUnicodeCharacter{255F}{+}",
            r"\DeclareUnicodeCharacter{254E}{|}",
            r"\DeclareUnicodeCharacter{257C}{-}",
            r"\DeclareUnicodeCharacter{257E}{-}",
            r"\DeclareUnicodeCharacter{2559}{+}",
        ]
    ),
}


# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "meds_tab_documentation.tex",
        "MEDS-TAB Documentation",
        r"Matthew McDermott \& Nassim Oufattole \& Teya Bergamaschi",
        "manual",
    )
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = ""

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# -- Options for EPUB output
epub_show_urls = "footnote"

print(f"loading configurations for {project} {release} ...", file=sys.stderr)


def setup(app):
    app.connect("builder-inited", ensure_pandoc_installed)



import json
from importlib.resources import files
from pathlib import Path

import hydra
from loguru import logger
from omegaconf import DictConfig, OmegaConf

from MEDS_tabular_automl.base_model import BaseModel

from ..utils import hydra_loguru_init, stage_init

config_yaml = files("MEDS_tabular_automl").joinpath("configs/launch_model.yaml")
if not config_yaml.is_file():
    raise FileNotFoundError("Core configuration not successfully installed!")


@hydra.main(version_base=None, config_path=str(config_yaml.parent.resolve()), config_name=config_yaml.stem)
def main(cfg: DictConfig) -> float:
    """Optimizes the model based on the provided configuration.

    Args:
        cfg: The configuration dictionary specifying model and training parameters.

    Returns:
        The evaluation result as the ROC AUC score on the held-out test set.
    """
    stage_init(
        cfg, ["input_dir", "input_label_cache_dir", "output_dir", "tabularization.filtered_code_metadata_fp"]
    )

    if not cfg.loguru_init:
        hydra_loguru_init()

    try:
        cfg.tabularization._resolved_codes
    except ValueError as e:
        logger.warning(f"No codes meet loading criteria, trial returning 0 AUC: {str(e)}")
        return 0.0

    model_launcher: BaseModel = hydra.utils.instantiate(cfg.model_launcher)

    model_launcher.train()
    auc = model_launcher.evaluate()

    # Make output model directory
    path_cfg = model_launcher.cfg.path
    model_filename = f"{path_cfg.model_file_stem}{path_cfg.model_file_extension}"
    model_config_hash = abs(hash(json.dumps(OmegaConf.to_container(cfg), sort_keys=True)))
    trial_output_dir = Path(path_cfg.sweep_results_dir) / str(model_config_hash)
    trial_output_dir.mkdir(parents=True, exist_ok=True)

    # save model
    model_launcher.save_model(trial_output_dir / model_filename)

    # save model config
    config_fp = trial_output_dir / f"{cfg.path.config_log_stem}.log"
    with open(config_fp, "w") as f:
        f.write(OmegaConf.to_yaml(cfg))

    # save model performance
    model_performance_fp = trial_output_dir / f"{cfg.path.performance_log_stem}.log"
    with open(model_performance_fp, "w") as f:
        f.write("trial_name,tuning_auc,test_auc\n")
        f.write(
            f"{trial_output_dir.stem},{model_launcher.evaluate()},"
            f"{model_launcher.evaluate(split='held_out')}\n"
        )

    logger.debug(f"Model config and performance logged to {config_fp} and {model_performance_fp}")
    return auc


if __name__ == "__main__":
    main()



import json
from importlib.resources import files
from pathlib import Path

import hydra
import pandas as pd
from loguru import logger
from omegaconf import DictConfig, OmegaConf

try:
    import autogluon.tabular as ag
except ImportError:
    ag = None

from MEDS_tabular_automl.tabular_dataset import TabularDataset as DenseIterator

from ..utils import hydra_loguru_init, stage_init

config_yaml = files("MEDS_tabular_automl").joinpath("configs/launch_model.yaml")
if not config_yaml.is_file():
    raise FileNotFoundError("Core configuration not successfully installed!")


def check_autogluon():
    if ag is None:
        raise ImportError(
            "AutoGluon could not be imported. Please try installing it using: `pip install autogluon`"
        )


@hydra.main(version_base=None, config_path=str(config_yaml.parent.resolve()), config_name=config_yaml.stem)
def main(cfg: DictConfig) -> float:
    """Launches AutoGluon after collecting data based on the provided configuration.

    Args:
        cfg: The configuration dictionary specifying model and training parameters.
    """
    check_autogluon()
    stage_init(
        cfg, ["input_dir", "input_label_cache_dir", "output_dir", "tabularization.filtered_code_metadata_fp"]
    )
    if not cfg.loguru_init:
        hydra_loguru_init()

    # collect data based on the configuration
    itrain = DenseIterator(cfg, "train")
    ituning = DenseIterator(cfg, "tuning")
    iheld_out = DenseIterator(cfg, "held_out")

    # collect data for AutoGluon
    train_data, train_labels = itrain.densify()
    tuning_data, tuning_labels = ituning.densify()
    held_out_data, held_out_labels = iheld_out.densify()

    # construct dfs for AutoGluon
    train_df = pd.DataFrame(train_data.todense())
    train_df[cfg.task_name] = train_labels
    tuning_df = pd.DataFrame(
        tuning_data.todense(),
    )
    tuning_df[cfg.task_name] = tuning_labels
    held_out_df = pd.DataFrame(held_out_data.todense())
    held_out_df[cfg.task_name] = held_out_labels

    train_dataset = ag.TabularDataset(train_df)
    tuning_dataset = ag.TabularDataset(tuning_df)
    held_out_dataset = ag.TabularDataset(held_out_df)

    # train model with AutoGluon
    log_filepath = Path(cfg.path.sweep_results_dir) / f"{cfg.path.config_log_stem}_log.txt"

    predictor = ag.TabularPredictor(
        label=cfg.task_name,
        log_to_file=True,
        log_file_path=str(log_filepath.resolve()),
        path=cfg.time_output_model_dir,
    ).fit(train_data=train_dataset, tuning_data=tuning_dataset)

    # predict
    predictions = predictor.predict(held_out_dataset.drop(columns=[cfg.task_name]))
    logger.info("Predictions:", predictions)
    # evaluate
    score = predictor.evaluate(held_out_dataset)
    logger.info("Test score:", score)

    model_performance_log_filepath = (
        Path(cfg.path.sweep_results_dir) / f"{cfg.path.performance_log_stem}.json"
    )
    model_performance_log_filepath.parent.mkdir(parents=True, exist_ok=True)
    # store results
    performance_dict = {
        "output_model_dir": cfg.path.time_output_model_dir,
        "tabularization": OmegaConf.to_container(cfg.tabularization),
        "model_launcher": OmegaConf.to_container(cfg.model_launcher),
        "score": score,
    }
    with open(model_performance_log_filepath, "w") as f:
        json.dump(performance_dict, f)


if __name__ == "__main__":
    main()



from collections.abc import Callable
from pathlib import Path

import scipy.sparse as sp
import xgboost as xgb
from loguru import logger
from omegaconf import DictConfig, OmegaConf
from sklearn.metrics import roc_auc_score

from .base_model import BaseModel
from .tabular_dataset import TabularDataset


class XGBIterator(xgb.DataIter, TabularDataset):
    """XGBIterator class for loading and processing data shards for use in XGBoost models.

    This class provides functionality for iterating through data shards, loading
    feature data and labels, and processing them based on the provided configuration.

    Args:
        cfg: A configuration dictionary containing parameters for
            data processing, feature selection, and other settings.
        split: The data split to use, which can be one of "train", "tuning",
            or "held_out". This determines which subset of the data is loaded and processed.

    Attributes:
        cfg: Configuration dictionary containing parameters for
            data processing, feature selection, and other settings.
        file_name_resolver: Object for resolving file names and paths based on the configuration.
        split: The data split being used for loading and processing data shards.
        _data_shards: List of data shard names.
        valid_event_ids: Dictionary mapping shard number to a list of valid event IDs.
        labels: Dictionary mapping shard number to a list of labels for the corresponding event IDs.
        codes_set: Set of codes to include in the data.
        code_masks: Dictionary of code masks for filtering features based on aggregation.
        num_features: Total number of features in the data.
    """

    def __init__(self, cfg: DictConfig, split: str):
        """Initializes the XGBIterator with the provided configuration and data split.

        Args:
            cfg: The configuration dictionary.
            split: The data split to use.
        """
        xgb.DataIter.__init__(self, cache_prefix=cfg.path.cache_dir)
        TabularDataset.__init__(self, cfg=cfg, split=split)

        self._it = 0

    def next(self, input_data: Callable) -> int:
        """Advances the XGBIterator by one step and provides data to XGBoost for DMatrix construction.

        Args:
            input_data: A function passed by XGBoost with the same signature as `DMatrix`.

        Returns:
            0 if end of iteration, 1 otherwise.
        """
        if self._it == len(self._data_shards):
            return 0

        X, y = self._get_shard_by_index(self._it)  # self._data_shards[self._it])
        logger.debug(f"X shape: {X.shape}, y shape: {y.shape}")
        input_data(data=sp.csr_matrix(X), label=y)
        self._it += 1

        return 1

    def reset(self):
        """Resets the XGBIterator to its beginning."""
        self._it = 0


class XGBoostModel(BaseModel):
    """Class for configuring, training, and evaluating an XGBoost model.

    This class utilizes the configuration settings provided to manage the training and evaluation
    process of an XGBoost model, ensuring the model is trained and validated using specified parameters
    and data splits. It supports training with in-memory data handling as well as direct streaming from
    disk using XGBIterators.

    Args:
        cfg: The configuration settings for the model, including data paths, model parameters,
            and flags for data handling.

    Attributes:
        cfg: Configuration object containing all settings required for model operation.
        model: The XGBoost model after being trained.
        dtrain: The training dataset in DMatrix format.
        dtuning: The tuning (validation) dataset in DMatrix format.
        dheld_out: The held-out (test) dataset in DMatrix format.
        itrain: XGBIterator for the training dataset.
        ituning: XGBIterator for the tuning dataset.
        iheld_out: XGBIterator for the held-out dataset.
        keep_data_in_memory: Flag indicating whether to keep all data in memory or stream from disk.
    """

    def __init__(self, cfg: DictConfig):
        """Initializes the XGBoostClassifier with the provided configuration.

        Args:
            cfg: The configuration dictionary.
        """
        super().__init__()
        self.cfg = cfg
        self.keep_data_in_memory = cfg.data_loading_params.keep_data_in_memory

        self.itrain = None
        self.ituning = None
        self.iheld_out = None

        self.dtrain = None
        self.dtuning = None
        self.dheld_out = None

        self.model = None

    def _build(self):
        """Builds necessary data structures for training."""
        if self.keep_data_in_memory:
            self._build_iterators()
            self._build_dmatrix_in_memory()
        else:
            self._build_iterators()
            self._build_dmatrix_from_iterators()

    def _train(self):
        """Trains the model."""
        self.model = xgb.train(
            OmegaConf.to_container(self.cfg.model),
            self.dtrain,
            num_boost_round=self.cfg.training_params.num_boost_round,
            early_stopping_rounds=self.cfg.training_params.early_stopping_rounds,
            evals=[(self.dtrain, "train"), (self.dtuning, "tuning")],
            verbose_eval=0,
        )

    def train(self):
        """Trains the model."""
        self._build()
        self._train()

    def _build_dmatrix_in_memory(self):
        """Builds the DMatrix from the data in memory."""
        X_train, y_train = self.itrain.get_data()
        X_tuning, y_tuning = self.ituning.get_data()
        X_held_out, y_held_out = self.iheld_out.get_data()
        self.dtrain = xgb.DMatrix(X_train, label=y_train)
        self.dtuning = xgb.DMatrix(X_tuning, label=y_tuning)
        self.dheld_out = xgb.DMatrix(X_held_out, label=y_held_out)

    def _build_dmatrix_from_iterators(self):
        """Builds the DMatrix from the iterators."""
        self.dtrain = xgb.DMatrix(self.itrain)
        self.dtuning = xgb.DMatrix(self.ituning)
        self.dheld_out = xgb.DMatrix(self.iheld_out)

    def _build_iterators(self):
        """Builds the iterators for training, validation, and testing."""
        self.itrain = XGBIterator(self.cfg, split="train")
        self.ituning = XGBIterator(self.cfg, split="tuning")
        self.iheld_out = XGBIterator(self.cfg, split="held_out")

    def evaluate(self, split="tuning") -> float:
        """Evaluates the model on the tuning set.

        Returns:
            The evaluation metric as the ROC AUC score.
        """
        if split == "tuning":
            y_pred = self.model.predict(self.dtuning)
            y_true = self.dtuning.get_label()
        elif split == "held_out":
            y_pred = self.model.predict(self.dheld_out)
            y_true = self.dheld_out.get_label()
        elif split == "train":
            y_pred = self.model.predict(self.dtrain)
            y_true = self.dtrain.get_label()
        else:
            raise ValueError(f"Invalid split for evaluation: {split}")
        return roc_auc_score(y_true, y_pred)

    def save_model(self, output_fp: Path):
        """Saves the trained model to the specified file path.

        Args:
            output_fp: The file path to save the model to.
        """
        self.model.save_model(output_fp)
